/******************************************************************************
 * bupman.cpp
 * Secondary wrapper around BupBoop to automatically manage resources.
 *-----------------------------------------------------------------------------
 * Copyright (C) 2015 - 2016 Osman Celimli
 * For conditions of distribution and use, see copyright notice in playtone.cpp
 ******************************************************************************/
#include <Windows.h>

#include <stdint.h>
#include <stdio.h>

#include "playtone.hpp"

/******************************************************************************
 * !!!!----               CONSTRUCTOR AND DESTRUCTOR                 ----!!!!
 ******************************************************************************/
BupMan::BupMan(HWND hParent)
{
	mhParent = hParent;

	mbRunning = FALSE;
	mbLogging = FALSE;

	mbGotSamples = FALSE;
	mbGotMacros = FALSE;
	mbGotMusic = FALSE;
	mbGotSFX = FALSE;

	mpSamples = NULL;
	mpMacros = NULL;
	mpMusic = NULL;
	mpSFX = NULL;
	miPriSFX = 0;
}

BupMan::~BupMan(void)
{
	stopDriver();

	if(mbGotSFX && (NULL != mpSFX))
	{
		delete[] mpSFX;
	}
	mbGotSFX = FALSE;
	mpSFX = NULL;

	if(mbGotMusic && (NULL != mpMusic))
	{
		delete[] mpMusic;
	}
	mbGotMusic = FALSE;
	mpMusic = NULL;

	if(mbGotMacros && (NULL != mpMacros))
	{
		delete[] mpMacros;
	}
	mbGotMacros = FALSE;
	mpMacros = NULL;

	if(mbGotSamples && (NULL != mpSamples))
	{
		delete[] mpSamples;
	}
	mbGotSamples = FALSE;
	mpSamples = NULL;
}




/******************************************************************************
 * !!!!----                  WRAPPED BUPBOOP CALLS                   ----!!!!
 ******************************************************************************/
/*  Each of these are wrapped around an equivalent BupBoop call with additional
 * protection against use without loading the proper consumables. See bupboop.h
 * for details.
 *----------------------------------------------------------------------------*/
void BupMan::pause(void)
{
	if(mbRunning)
	{
		bb_pause();
	}
}

void BupMan::resume(void)
{
	if(mbRunning)
	{
		bb_resume();
	}
}

INT BupMan::isPaused(void)
{
	INT iPaused = 0;

	if(mbRunning)
	{
		iPaused = bb_isPaused();
	}

	return iPaused;
}

void BupMan::stopAll(void)
{
	if(mbRunning)
	{
		bb_stopAll();
	}
}




void BupMan::playMusic(void)
{
	if(mbRunning && mbGotMusic)
	{
		bb_playMusic(mpMusic);
	}
}

void BupMan::stopMusic(void)
{
	if(mbRunning)
	{
		bb_stopMusic();
	}
}

void BupMan::attenMusic(BYTE cVol)
{
	if(mbRunning)
	{
		bb_attenMusic(cVol);
	}
}

INT BupMan::checkMusic(void)
{
	INT iStatus = 0;

	if(mbRunning)
	{
		iStatus = bb_checkMusic();
	}

	return iStatus;
}

INT BupMan::getMood(void)
{
	INT iMood = 0;

	if(mbRunning)
	{
		iMood = bb_getMood();
	}

	return iMood;
}




void BupMan::playSFX(INT iPriority, INT iVol_Left, INT iVol_Right)
{
	if(mbRunning && mbGotSFX && (0 != iPriority))
	{
		miPriSFX = iPriority;
		bb_playSFX(mpSFX, miPriSFX, iVol_Left, iVol_Right);
	}
}

void BupMan::stopSFX(void)
{
	if(0 != miPriSFX)
	{
		bb_stopSFX(miPriSFX);
	}
}




/******************************************************************************
 * !!!!----                     DRIVER CONTROL                       ----!!!!
 ******************************************************************************/
/* HRESULT BupMan::startDriver(void)
 *  Attempt to start running BupBoop if and only if the sample and macro packs
 * have been loaded and BupBoop is not already running.
 *----------------------------------------------------------------------------*/
HRESULT BupMan::startDriver(void)
{
	HRESULT hRes = E_FAIL;

	if(!mbRunning && mbGotSamples && mbGotMacros)
	{
		hRes = wt_init(mhParent, mpSamples, mpMacros);
		mbRunning = FAILED(hRes)
			? FALSE
			: TRUE;
	}

	return hRes;
}

/* HRESULT BupMan::stopDriver(void)
 *  Close down BupBoop if it is currently running, this must be performed
 * prior to changing sample or instrument sets. If a RIFF/WAVE log is currently
 * running it will be closed as well.
 *----------------------------------------------------------------------------*/
HRESULT BupMan::stopDriver(void)
{
	HRESULT hRes = E_FAIL;

	if(mbRunning)
	{
		stopLog();

		bb_stopAll();
		wt_close();

		mbRunning = FALSE;
		mbLogging = FALSE;
		hRes = S_OK;
	}

	return hRes;
}




/* HRESULT BupMan::startLog(LPSTR pszFile)
 *----------------------------------------------------------------------------*/
HRESULT BupMan::startLog(LPSTR pszFile)
{
	HRESULT hRes = E_FAIL;

	if(mbRunning && !mbLogging)
	{
		if(0 == wt_logOutput(pszFile))
		{
			mbLogging = TRUE;
			hRes = S_OK;
		}
	}

	return hRes;
}

/* HRESULT BupMan::stopLog(void)
 *----------------------------------------------------------------------------*/
HRESULT BupMan::stopLog(void)
{
	HRESULT hRes = E_FAIL;

	if(mbRunning && mbLogging)
	{
		wt_closeLog();

		mbLogging = FALSE;
		hRes = S_OK;
	}

	return hRes;
}




/* HRESULT BupMan::loadFile(LPSTR pszFile, BYTE **ppDest)
 *  General use binary file loader, will attempt to load the file whose name
 * is in pszFile to a new'd block of memory whose base address will be written
 * to *ppDest.
 *----------------------------------------------------------------------------*/
HRESULT BupMan::loadFile(LPSTR pszFile, BYTE **ppDest)
{
	HRESULT hRes = E_ACCESSDENIED;

	FILE *pFile = NULL;
	BYTE *pData = NULL;

	UINT uiLen,uiRead;

	if(NULL != (pFile = fopen(pszFile, "rb")))
	{
		fseek(pFile, 0, SEEK_END);
		uiLen = ftell(pFile);
		rewind(pFile);

		pData = new BYTE[uiLen];
		if(NULL == pData)
		{
			hRes = E_OUTOFMEMORY;
		}
		else
		{
			uiRead = fread(pData, sizeof(char), uiLen, pFile);
			fclose(pFile);

			if(uiRead == uiLen)
			{
				*ppDest = pData;
				hRes = S_OK;
			}
			else
			{
				delete[] pData;
				hRes = E_UNEXPECTED;
			}
		}
	}

	return hRes;
}

/* HRESULT BupMan::loadSamples(LPSTR pszFile)
 *  Attempt to load the sample package whose filename is in pszFile, note that
 * this will automatically close the driver if it is currently running.
 *----------------------------------------------------------------------------*/
HRESULT BupMan::loadSamples(LPSTR pszFile)
{
	HRESULT hRes = E_FAIL;

	if(mbRunning)
	{
		stopDriver();
	}

	if(mbGotSamples)
	{
		delete[] mpSamples;

		mpSamples = NULL;
		mbGotSamples = FALSE;
	}

	hRes = loadFile(pszFile, &mpSamples);
	mbGotSamples = FAILED(hRes)
		? FALSE
		: TRUE;

	return hRes;
}

/* HRESULT BupMan::loadMacros(LPSTR pszFile)
 *  Attempt to load the macro package whose filename is in pszFile, note that
 * this will automatically close the driver if it is currently running.
 *----------------------------------------------------------------------------*/
HRESULT BupMan::loadMacros(LPSTR pszFile)
{
	HRESULT hRes = E_FAIL;

	if(mbRunning)
	{
		stopDriver();
	}

	if(mbGotMacros)
	{
		delete[] mpMacros;

		mpMacros = NULL;
		mbGotMacros = FALSE;
	}

	hRes = loadFile(pszFile, &mpMacros);
	mbGotMacros = FAILED(hRes)
		? FALSE
		: TRUE;

	return hRes;
}

/* HRESULT BupMan::loadMusic(LPSTR pszFile)
 *  Attempt to load the music track whose filename is in pszFile, this will
 * automatically cease playback of any currently active music.
 *----------------------------------------------------------------------------*/
HRESULT BupMan::loadMusic(LPSTR pszFile)
{
	HRESULT hRes = E_FAIL;

	if(mbGotMusic)
	{
		stopMusic();
		delete[] mpMusic;

		mpMusic = NULL;
		mbGotMusic = FALSE;
	}

	hRes = loadFile(pszFile, &mpMusic);
	mbGotMusic = FAILED(hRes)
		? FALSE
		: TRUE;

	return hRes;
}

/* HRESULT BupMan::loadSFX(LPSTR pszFile)
 *  Attempt to load the sound effect whose filename is in pszFile, this will
 * automatically cease playback of any currently active sound effect.
 *----------------------------------------------------------------------------*/
HRESULT BupMan::loadSFX(LPSTR pszFile)
{
	HRESULT hRes = E_FAIL;

	if(mbGotSFX)
	{
		/**
		 * All sound effects dispatched from the MENU have a default priority
		 * of 127 and full volume center panning. However, using the sound
		 * effect manager to dispatch more instances of the currently loaded
		 * effect will overwrite the cached priority from the menu.
		 *
		 * So before we load, we'll stop both the sound effect with the last
		 * logged priority and the menu priority of 127. This isn't an issue
		 * for game software, which is expected to have all sound effects
		 * loaded up front.
		 */
		stopSFX();
		miPriSFX = 127;
		stopSFX();

		delete[] mpSFX;

		mpSFX = NULL;
		mbGotSFX = FALSE;
	}

	hRes = loadFile(pszFile, &mpSFX);
	mbGotSFX = FAILED(hRes)
		? FALSE
		: TRUE;

	return hRes;
}
