/******************************************************************************
 * playtone.cpp
 * CoreTone Auditioning Suite for Windows.
 *-----------------------------------------------------------------------------
 * Version 1.2.2cz, November 26th, 2016
 * Copyright (C) 2015 - 2016 Osman Celimli
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 ******************************************************************************/
#include <Windows.h>
#include <Commctrl.h>
#include <tchar.h>

#include <stdint.h>

#include "playtone.hpp"

/******************************************************************************
 * !!!!----                   WINDOWS GUI GIBBUMS                    ----!!!!
 ******************************************************************************/
/* LoadString Buffer
 *-----------------------------------*/
#define STRING_BUFFER_ENTRIES	512
#define STRING_BUFFER_LEN		(sizeof(szBuffer) / sizeof(szBuffer[0]))

TCHAR szBuffer[STRING_BUFFER_ENTRIES];


/* Main Window and Friends
 *-----------------------------------*/
const TCHAR szWinClass[] = _T("ToneClass");
const TCHAR szWinName[] = _T("PlayTone");

const TCHAR szEmpty[] = _T("");
const TCHAR szTerms[] = _T(":\\/>");
WNDCLASSEX wcWinMain;

HWND hMainWnd = NULL;
UINT_PTR pDrawTimer = NULL;

BOOL bFocus = false,bMinimized = false;

BupMan *pBupMan = NULL;

MenuMan *pMenuMan = NULL;
WaveMan *pWaveMan = NULL;
StatMan *pStatMan = NULL;
PanelMan *pPanelMan = NULL;

PatchPop *pPatchPop = NULL;
MusicPop *pMusicPop = NULL;
EffectPop *pEffectPop = NULL;


/******************************************************************************
 * !!!!----                   WINMAIN and WINPROC                    ----!!!!
 ******************************************************************************/
/* static LRESULT CALLBACK WinProc(HWND hWnd, UINT uMsg,
 *                                 WPARAM wParam, LPARAM lParam)
 * Window process...
 *----------------------------------------------------------------------------*/
static LRESULT CALLBACK WinProc(HWND hWnd, UINT uMsg,
                                WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT psWin;
	HDC hDC;

	LRESULT lRes;

	switch(uMsg)
	{
		/**
		 * ---- WM_CREATE ----
		 * The window is currently being instantiated, i.e. this is
		 * encountered while still in CreateWindow / CreateWindowEx.
		 */
		case WM_CREATE:
			pBupMan = new BupMan(hWnd);
			pMenuMan = new MenuMan(hWnd);

			pWaveMan = new WaveMan(hWnd, MAIN_WAVE_XPOS, MAIN_WAVE_YPOS,
				MAIN_WAVE_WIDTH, MAIN_WAVE_HEIGHT);
			pStatMan = new StatMan(hWnd, MAIN_STAT_XPOS, MAIN_STAT_YPOS,
				MAIN_STAT_WIDTH, MAIN_STAT_HEIGHT);
			pPanelMan = new PanelMan(hWnd, MAIN_PANEL_XPOS, MAIN_PANEL_YPOS);

			if((NULL == pBupMan) || (NULL == pMenuMan) || (NULL == pWaveMan)
				|| (NULL == pStatMan) || (NULL == pPanelMan))
			{
				DestroyWindow(hWnd);
				break;
			}
			break;

		/**
		 * ---- WM_DESTROY ----
		 * The window is being destroyed. Note that we'll only receive
		 * this message once all of this window's children have already
		 * been synchronously destroyed.
		 *
		 * This means that once we post the quit messaage to WinMain's
		 * message loop, it is guaranteed that no windows are active.
		 */
		case WM_DESTROY:
			if(NULL != pDrawTimer)
			{
				KillTimer(hWnd, DRAW_TIMER_ID);
				pDrawTimer = NULL;
			}

			PostQuitMessage(0);
			break;

		/**
		 * ---- WM_CLOSE ----
		 * The application is being terminated, this is the message
		 * received when the user clicks the little "X".
		 */
		case WM_CLOSE:
			DestroyWindow(hWnd);
			break;


		/**
		 * ---- WM_TIMER ----
		 * The timer event associated with this window has expired, we should
		 * redraw if BupBoop is running and we're not minimized.
		 */
		case WM_TIMER:
			if(pBupMan->mbRunning && !bMinimized)
			{
				UpdateWindow(hWnd);

				pWaveMan->invalProc(hWnd);
				pPanelMan->refresh();

				if(NULL != pPatchPop)
				{
					if(pPatchPop->checkPop())
					{
						pPatchPop->pushList();
					}
				}

				if(NULL != pMusicPop)
				{
					if(pMusicPop->checkPop())
					{
						pMusicPop->pushList();
					}
				}
			}
			break;


		/**
		 * ---- WM_ERASEBKGND ----
		 * The window background must be erased (and the window
		 * has possibly been resized). We should return a nonzero
		 * value as long as the background is erased properly.
		 */
		case WM_ERASEBKGND:
			break;

		/**
		 * ---- WM_PAINT ----
		 * Time to redraw!
		 */
		case WM_PAINT:
			hDC = BeginPaint(hWnd, &psWin);
			if(NULL != hDC)
			{
				pWaveMan->paintProc(hWnd, hDC, psWin);
				pStatMan->paintProc(hWnd, hDC, psWin);
				EndPaint(hWnd, &psWin);
			}
			break;


		/**
		 * ---- WM_SIZING ----
		 * The window is being resized by the user, i.e. they're
		 * dragging one of its sides or corners. We can intercept
		 * and alter the new sizing parameters if desired.
		 *
		 * Note that this message will NOT be received when
		 * minimizing or maximizing, for that use WM_SIZE (below).
		 */
		case WM_SIZING:
			break;

		/**
		 * ---- WM_SETFOCUS, WM_KILLFOCUS, and WM_ACTIVATE ----
		 * We've either gained or lost focus due to the user
		 * interacting with another window or minimizing ours.
		 */
		case WM_SETFOCUS:
			bFocus = true;
			break;
		case WM_KILLFOCUS:
			bFocus = false;
			break;

		case WM_ACTIVATE:
			bFocus = (WA_ACTIVE == LOWORD(wParam));
			break;

		/**
		 * ---- WM_SIZE ----
		 * The Window has been resized, this includes maximizing
		 * or minimizing the window (in the latter case its width
		 * and height will be zero).
		 */
		case WM_SIZE:
			bMinimized = (SIZE_MINIMIZED == wParam)
				? true
				: false;

			if(!bMinimized)
			{
				pPanelMan->refresh();
			}
			break;

		/**
		 * ---- WM_ENTERSIZEMOVE and WM_EXITSIZEMOVE ----
		 * A resize modal loop has been entered or exited.
		 *
		 *                     ...or...
		 *
		 * ---- WM_ENTERMENULOOP and WM_EXITMENULOOP ----
		 * A menu modal loop has been entered or exited.
		 *
		 * These will block and unblock the message loop,
		 * respectively.
		 */
		case WM_ENTERSIZEMOVE:
		case WM_ENTERMENULOOP:
			break;
		case WM_EXITSIZEMOVE:
		case WM_EXITMENULOOP:
			break;


		/**
		 * ---- WM_COMMAND ----
		 * Any of the menu options (IDM_x) have been selected or
		 * an accelerator key combination has been pressed.
		 */
		case WM_COMMAND:
			lRes = pMenuMan->cmdProc(hWnd, uMsg, wParam, lParam);
			lRes |= pPanelMan->cmdProc(hWnd, uMsg, wParam, lParam);

			pPanelMan->refresh();
			return lRes;

		/**
		 * ---- WM_HSCROLL ----
		 * An interaction with a horizontal scrollbar or trackbar
		 * has been performed.
		 *
		 * wParam indicates the type of interaction while lParam
		 * is the handle of who was interacted with.
		 */
		case WM_HSCROLL:
			lRes = pPanelMan->trackProc(hWnd, uMsg, wParam, lParam);

			return lRes;


		/**
		 * How did we get here?
		 * Windows, figure it out for us.
		 */
		default:
			break;
	}

	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}




/* int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
 *                    LPSTR lpCmdLine, int nShowCmd)
 * WinMain...
 * It's Main for Wins... yeah.
 *----------------------------------------------------------------------------*/
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
				   LPSTR lpCmdLine, int nShowCmd)
{
	HICON hIcon32,hIcon16;
	HICON hCursor;

	HACCEL hAccels;
	MSG MainMsg;

	INT iX,iY;
	INT iStatus;


	/**
	 * ---- BUPBOOP SAFETY CHECK ----
	 * Ensure our loaded WinTone + CoreTone DLLs are a matched set (identical
	 * channel count, render frequency, etc) before starting the application.
	 */
	iStatus = wt_getInfo(NULL, NULL, NULL, NULL, NULL);

	if(iStatus)
	{
		iStatus = LoadString(hInstance, IDS_CORE_MISMATCH,
			szBuffer, STRING_BUFFER_LEN);
		if(iStatus > 0)
		{
			MessageBox(NULL, szBuffer, NULL, (MB_OK | MB_ICONEXCLAMATION));
		}

		goto cleanup;
	}


	/**
	 * Load up our icons and cursors, these will be automatically freed
	 * when this program terminates, but according to the MSDN it
	 * is good manners to release them using DestroyIcon / DestroyCursor.
	 *
	 * Note that since we're using the default system cursor,
	 * we shouldn't destroy it.
	 */
	hIcon32 = (HICON)LoadImage(
		hInstance,
		MAKEINTRESOURCE(IDI_CORETONE),
		IMAGE_ICON,
		32, 32,
		LR_DEFAULTCOLOR);

	hIcon16 = (HICON)LoadImage(
		hInstance,
		MAKEINTRESOURCE(IDI_CORETONE),
		IMAGE_ICON,
		16, 16,
		LR_DEFAULTCOLOR);

	hCursor = LoadCursor(NULL, IDC_ARROW);


	/**
	 * ---- Window Class Setup ----
	 * Every window has a window class which defines basic appearance
	 * and system behavior within it such as which mouse cursor to use.
	 *
	 * These all must be registered using RegisterClassEx prior to
	 * creating any actual windows.
	 */
	wcWinMain.cbSize = sizeof(WNDCLASSEX);
	wcWinMain.style = 0;
	wcWinMain.lpfnWndProc = WinProc;
	wcWinMain.cbClsExtra = 0;
	wcWinMain.cbWndExtra = 0;
	wcWinMain.hInstance = hInstance;
	wcWinMain.hIcon = hIcon32;
	wcWinMain.hCursor = hCursor;
	wcWinMain.hbrBackground = (HBRUSH)(COLOR_WINDOW + 0);
	wcWinMain.lpszMenuName = MAKEINTRESOURCE(IDR_MENU_MAIN);
	wcWinMain.lpszClassName = szWinClass;
	wcWinMain.hIconSm = hIcon16;

	if(!RegisterClassEx(&wcWinMain))
	{
		iStatus = LoadString(hInstance, IDS_ERR_WINCLASS,
			szBuffer, STRING_BUFFER_LEN);
		if(iStatus > 0)
		{
			MessageBox(NULL, szBuffer, NULL, (MB_OK | MB_ICONEXCLAMATION));
		}

		goto cleanup;
	}


	/**
	 * ---- Main Window Creation ----
	 * Now we can actually initialize our main window by supplying
	 * CreateWindowEx with a metric boatload of arguments.
	 */
	InitCommonControls();
	iX = MAIN_WINDOW_WIDTH;
	iY = MAIN_WINDOW_HEIGHT;

	hMainWnd = CreateWindowEx(
		(0),									/* dwExStyle */
		szWinClass,								/* lpClassName */
		szWinName,								/* lpWindowName */
		(WS_SYSMENU | WS_CAPTION | WS_VISIBLE
			| WS_MINIMIZEBOX),					/* dwStyle */
		CW_USEDEFAULT,							/* x */
		CW_USEDEFAULT,							/* y */
		iX,										/* nWidth */
		iY,										/* nHeight */
		NULL,									/* hWndParent */
		NULL,									/* hMenu, NULL = Use CLASS Menu */
		hInstance,								/* hInstance */
		NULL);									/* lpParam */

	if(NULL == hMainWnd)
	{
		iStatus = LoadString(hInstance, IDS_ERR_WINCREATE,
			szBuffer, STRING_BUFFER_LEN);
		if(iStatus > 0)
		{
			MessageBox(NULL, szBuffer, NULL, (MB_OK | MB_ICONEXCLAMATION));
		}

		goto cleanup;
	}


	/**
	 * ---- Message Loop ----
	 * Sit in our dispatch loop until the user quits or something
	 * self-destructs.
	 */
	pDrawTimer = SetTimer(hMainWnd, DRAW_TIMER_ID, DRAW_TIMER_MS, NULL);
	hAccels = LoadAccelerators(hInstance, MAKEINTRESOURCE(IDR_ACCELERATOR));

	while(GetMessage(&MainMsg, NULL, 0, 0) > 0)
	{
		if(WM_QUIT != MainMsg.message)
		{
			/* ++++ HANDLE MESSAGE ++++
			 */
			if(NULL != pPatchPop)
			{
				/* ???? PATCH DIALOG ????
				 */
				if(pPatchPop->checkPop()
					&& IsDialogMessage(pPatchPop->getPop(), &MainMsg))
				{
					continue;
				}
			}

			if(NULL != pMusicPop)
			{
				/* ???? MUSIC DIALOG ????
				 */
				if(pMusicPop->checkPop()
					&& IsDialogMessage(pMusicPop->getPop(), &MainMsg))
				{
					continue;
				}
			}

			if(NULL != pEffectPop)
			{
				/* ???? SOUND EFFECT DIALOG ????
				 */
				if(pEffectPop->checkPop()
					&& IsDialogMessage(pEffectPop->getPop(), &MainMsg))
				{
					continue;
				}
			}


			/* >>>> MAIN HWND >>>>
			 */
			if(!TranslateAccelerator(hMainWnd, hAccels, &MainMsg))
			{
				TranslateMessage(&MainMsg);
				DispatchMessage(&MainMsg);
			}
		}
		else
		{
			/* ++++ QUIT ++++
			 */
			break;
		}
	}


	/**
	 * ---- Exit ----
	 * Clean up and we're outta here!
	 */
cleanup:
	if(NULL != pDrawTimer)
	{
		KillTimer(hMainWnd, DRAW_TIMER_ID);
		pDrawTimer = NULL;
	}


	if(NULL != pPatchPop)
	{
		delete pPatchPop;
		pPatchPop = NULL;
	}

	if(NULL != pMusicPop)
	{
		delete pMusicPop;
		pMusicPop = NULL;
	}

	if(NULL != pEffectPop)
	{
		delete pEffectPop;
		pEffectPop = NULL;
	}


	if(NULL != pPanelMan)
	{
		delete pPanelMan;
		pPanelMan = NULL;
	}

	if(NULL != pStatMan)
	{
		delete pStatMan;
		pStatMan = NULL;
	}

	if(NULL != pWaveMan)
	{
		delete pWaveMan;
		pWaveMan = NULL;
	}

	if(NULL != pMenuMan)
	{
		delete pMenuMan;
		pMenuMan = NULL;
	}


	if(NULL != pBupMan)
	{
		pBupMan->stopDriver();

		delete pBupMan;
		pBupMan = NULL;
	}


	if(NULL != hIcon32)
		DestroyIcon(hIcon32);
	if(NULL != hIcon16)
		DestroyIcon(hIcon16);

	return MainMsg.wParam;
}
