/******************************************************************************
 * wintone.hpp
 * Windows + DirectSound wrappers for CoreTone.
 *-----------------------------------------------------------------------------
 * Copyright (C) 2015 - 2016 Osman Celimli
 * For conditions of distribution and use, see copyright notice in wintone.cpp
 ******************************************************************************/
#ifndef WINTONE
#define WINTONE
/******************************************************************************
 * Operating Parameters
 ******************************************************************************/
/* ---- A Note on Timing ----
 * CoreTone is designed to have a rendering frequency which is a multiple of
 * its script decoding frequency so it can perform both script and rendering
 * updates in a single step. For example, the default rendering frequency
 * of 48000Hz and decoding frequency of 240Hz results in a fixed 200 sample
 * pairs per update (48000Hz / 240Hz = 200).
 *
 * WinTone is just a wrapper around CoreTone to perform these updates as
 * close as possible to these desired update rates and shove the results into
 * DirectSound. If you absolutely require different timing, it may be best to
 * modify the #def'd parameters in CoreTone itself.
 *
 *
 * ---- Anyway, the defines.... ----
 * WINTONE_UPDATE_RATE is the rate in Hz at which WinTone will perform its
 * rendering update checks. For most use cases, this should be left equal to
 * CORETONE_DECODE_RATE and in all cases it should attempt to be an even
 * divisor of CORETONE_DECODE_RATE.
 *
 * WINTONE_UPDATE_TICKS is the maximum number of CoreTone update ticks to
 * perform in one WinTone update tick (which should be one or greater, larger
 * values will create a higher latency).
 *
 * WINTONE_RBUFFER_LEN is the length of the WinTone render buffer itself in
 * interleaved mono samples, this should always be CORETONE_BUFFER_LEN
 * multiplied by the number of update ticks, WINTONE_UPDATE_TICKS.
 *
 * WINTONE_DBUFFER_LEN is the length of the DirectSound buffer in samples,
 * which must be two or more times the size of the WinTone render buffer.
 */
#ifndef WINTONE_UPDATE_RATE
	#define WINTONE_UPDATE_RATE			(CORETONE_DECODE_RATE / 4)
#endif

#ifndef WINTONE_UPDATE_TICKS
	#define WINTONE_UPDATE_TICKS		(CORETONE_DECODE_RATE / WINTONE_UPDATE_RATE)
#endif

#ifndef WINTONE_RBUFFER_LEN
	#define WINTONE_RBUFFER_LEN			(CORETONE_BUFFER_LEN * WINTONE_UPDATE_TICKS)
#endif

#ifndef WINTONE_DBUFFER_LEN
	#define WINTONE_DBUFFER_LEN			(WINTONE_RBUFFER_LEN * 32)
#endif

#ifdef WINTONE_EXPORTS
	#define WINTONE_API __declspec(dllexport)
#else
	#define WINTONE_API __declspec(dllimport)
#endif


/* WINTONE_WRITE_BORDER_START is the minimum distance we're placed ahead of
 * the write cursor on the first render tick.
 *
 * WINTONE_WRITE_BORDER_MAX is the maximum distance the render offset can drift
 * ahead of the write cursor (due to missed updates).
 *
 * WINTONE_PUSH_OFFSET is the minimum distance we must stay ahead of the write
 * cursor before we try to bump the render offset up to the current write border.
 */
#ifndef WINTONE_WRITE_BORDER_START
	#define WINTONE_WRITE_BORDER_START	4
#endif

#ifndef WINTONE_WRITE_BORDER_MAX
	#define WINTONE_WRITE_BORDER_MAX	16
#endif

#ifndef WINTONE_PUSH_OFFSET
	#define WINTONE_PUSH_OFFSET			2
#endif


/* WINTONE_UPDATE_LATENCY_MS is how often (in milliseconds) WinTone will check
 * on DirectSound's playback of CoreTone's rendered output in order to determine
 * if it's time to start the next render cycle.
 *
 * WINTONE_UPDATE_RESOLUTION_MS is the accuracy (in milliseconds) of the timer
 * used to check the condition above.
 */
#ifndef WINTONE_UPDATE_LATENCY_MS
	#define WINTONE_UPDATE_LATENCY_MS		(1000 / (WINTONE_UPDATE_RATE * 4))
#endif

#ifndef WINTONE_UPDATE_RESOLUTION_MS
	#define WINTONE_UPDATE_RESOLUTION_MS	2
#endif


/* Identical to ct_renderCall_t aside from the name.
 */
typedef int32_t (*wt_renderCall_t)(void *pBuffer,
								   uint32_t uiFreq, uint32_t uiLen,
								   int32_t iAmPaused);


/******************************************************************************
 * Function Defines
 ******************************************************************************/
WINTONE_API void wt_pause(void);
WINTONE_API void wt_resume(void);
WINTONE_API int32_t wt_isPaused(void);
WINTONE_API void wt_stopAll(void);
WINTONE_API void wt_setRenderCall(wt_renderCall_t pCall);

WINTONE_API void wt_playMusic(uint8_t *pMusic);
WINTONE_API void wt_stopMusic(void);
WINTONE_API void wt_attenMusic(int8_t cVol);
WINTONE_API int32_t wt_checkMusic(void);
WINTONE_API int32_t wt_getMood(void);

WINTONE_API void wt_playSFX(uint8_t *pSFX, int8_t cPriority, int8_t cVol_Left, int8_t cVol_Right);
WINTONE_API void wt_stopSFX(int8_t cPriority);
WINTONE_API void wt_addSFX(uint8_t *pSFX, int8_t cPriority, int8_t cVol_Left, int8_t cVol_Right);
WINTONE_API void wt_dumpSFX(void);

WINTONE_API int32_t wt_logOutput(char *pszFile);
WINTONE_API void wt_closeLog(void);

WINTONE_API int32_t wt_getMutex(void);
WINTONE_API int32_t wt_giveMutex(void);

#ifndef BUPBOOP_EXCLUDE_OS_SPECIFIC
WINTONE_API HRESULT wt_init(HWND hParent, uint8_t *pSamplePak, uint8_t *pInstrPak);
WINTONE_API void wt_close(void);
#endif

WINTONE_API void wt_getState(void **ppaChannels, void **ppaPatches, void **ppaTracks,
                             int32_t *piToneReady, uint32_t *puiTicks,
                             int16_t **ppaRenderWave, uint32_t *puiRenderLen);

WINTONE_API int32_t wt_getInfo(uint32_t *puiChannels, uint32_t *puiRenderFreq,
                               uint32_t *puiDecodeRate,
                               uint32_t *puiSamples, uint32_t *puiSampleLen);
#endif
